import pykube
from yaml import load
from os.path import expanduser

from bgkube.utils import dot_env_dict, read_with_merge_vars


class KubeApi:
    def __init__(self):
        self._client = None

    def client(self):
        if not self._client:
            self._client = pykube.HTTPClient(pykube.KubeConfig.from_file(expanduser('~/.kube/config')))
        return self._client

    def get_config_with_vars(self, config_file, env_file, **attrs):
        merge_vars = dict(dot_env_dict(env_file)) if env_file else {}
        merge_vars.update(attrs)

        for section in read_with_merge_vars(config_file, merge_vars).split('---'):
            yield load(section)

    def apply(self, config_file, env_file, **attrs):
        def apply_object(config):
            obj = getattr(pykube, config['kind'])(self.client(), config)

            if obj.exists():
                obj.update()
            else:
                obj.create()

            return obj.name

        objects = [apply_object(config) for config in self.get_config_with_vars(config_file, env_file, **attrs)]
        return objects

    def resource_by_name(self, resource, name):
        try:
            return getattr(pykube, resource).objects(self.client()).get_by_name(name)
        except pykube.ObjectDoesNotExist:
            return None

    def pods(self, **labels):
        return list(pykube.Pod.objects(self.client()).filter(selector=labels))



from six import add_metaclass
from time import sleep

from bgkube import cmd, registries
from bgkube.api import KubeApi
from bgkube.run import Runner
from bgkube.errors import ActionFailedError
from bgkube.utils import output, log, timestamp, require, get_loadbalancer_address, is_host_up


class BgKubeMeta(type):
    required = [
        'cluster_zone', 'cluster_name', 'image_name', 'service_name', 'service_config', 'deployment_config'
    ]
    optional = [
        'context', 'dockerfile', 'env_file', 'smoke_tests_command', 'smoke_service_config', 'docker_machine_name',
        'db_migrations_job_config_seed', 'db_migrations_status_command', 'db_migrations_apply_command',
        'db_migrations_rollback_command', 'kops_state_store', 'container_registry', 'service_timeout',
        'smoke_service_timeout', 'deployment_timeout', 'db_migrations_job_timeout', 'docker_build_args'
    ]
    optional_defaults = {
        'context': '.',
        'dockerfile': './Dockerfile',
        'container_registry': registries.DEFAULT,
        'service_timeout': 120,
        'smoke_service_timeout': 120,
        'deployment_timeout': 120,
        'db_migrations_job_timeout': 120,
        'docker_build_args': ''
    }

    def __new__(mcs, name, bases, attrs):
        attrs['required'] = mcs.required
        attrs['optional'] = mcs.optional

        for field in mcs.required + mcs.optional:
            attrs[field] = mcs.optional_defaults.get(field, None)

        return super(BgKubeMeta, mcs).__new__(mcs, name, bases, attrs)


@add_metaclass(BgKubeMeta)
class BgKube(object):
    def __init__(self, options):
        self.load_options(options)

        self.kube_api = KubeApi()
        self.runner = Runner(self.get_docker_daemon())
        self.registry = registries.load(self.runner, options)

    @property
    def is_minikube(self):
        return self.container_registry == 'local'

    def get_docker_daemon(self):
        return cmd.MINIKUBE_DOCKER_ENV if self.is_minikube else cmd.DOCKERMACHINE_ENV.format(self.docker_machine_name)

    def load_options(self, options):
        for opt in self.required:
            setattr(self, opt, require(options, opt))

        for opt in self.optional:
            setattr(self, opt, getattr(options, opt, None) or getattr(self, opt))

    @log('Building image {image_name} using {dockerfile}...')
    def build(self):
        tag = timestamp()
        command = [cmd.DOCKER_BUILD.format(
            context=self.context,
            dockerfile=self.dockerfile,
            image=self.image_name,
            tag=tag,
        )]

        if self.docker_build_args:
            command.append(' '.join('--build-arg {}'.format(b) for b in self.docker_build_args.split(' ')))

        self.runner.start(' '.join(command))
        return tag

    @log('Pushing image {image_name}:{tag} to {registry}...')
    def push(self, tag):
        self.registry.push('{}:{}'.format(self.image_name, tag))

    @log('Applying {_} using config: {filename}...')
    def apply(self, _, filename, tag=None, color=''):
        return self.kube_api.apply(filename, self.env_file, TAG=tag, COLOR=color, ENV_FILE=self.env_file)

    def pod_find(self, tag, color):
        results = [pod for pod in self.kube_api.pods(tag=tag, color=color) if pod.ready]
        return results[0] if results else None

    def pod_exec(self, tag, color, command, *args):
        pod = self.pod_find(tag, color).name
        return self.runner.start(cmd.KUBECTL_EXEC.format(pod=pod, command=command, args=' '.join(args)), capture=True)

    def migrate_initial(self, tag):
        if self.db_migrations_job_config_seed:
            def job_completions_extractor(job):
                completions = job.obj['spec']['completions']
                succeeded_completions = job.obj['status']['succeeded']

                return completions if succeeded_completions == completions else None

            applied_objects = self.apply('db migration', self.db_migrations_job_config_seed, tag=tag)
            self.wait_for_resource_running(
                'Job',
                'completions',
                job_completions_extractor,
                self.db_migrations_job_timeout,
                *applied_objects
            )

    def migrate_apply(self, tag, color):
        previous_state = None

        if self.db_migrations_status_command:
            previous_state = self.pod_exec(tag, color, self.db_migrations_status_command)

        if self.db_migrations_apply_command:
            self.pod_exec(tag, color, self.db_migrations_apply_command)

        return previous_state

    def migrate_rollback(self, tag, color, previous_state):
        if self.db_migrations_rollback_command:
            self.pod_exec(tag, color, self.db_migrations_rollback_command, previous_state)

    def migrate(self, tag, color):
        db_migrations_previous_state = None
        is_initial = self.active_env() is None

        if is_initial:
            self.migrate_initial(tag)
        else:
            db_migrations_previous_state = self.migrate_apply(tag, color)

        return is_initial, db_migrations_previous_state

    def active_env(self):
        service = self.kube_api.resource_by_name('Service', self.service_name)
        return None if not service else service.obj['spec']['selector']['color']

    def other_env(self):
        return {
            'blue': 'green',
            'green': 'blue'
        }.get(self.active_env(), None)

    def deploy(self, tag):
        color = self.other_env() or 'blue'
        applied_objects = self.apply('deployment', self.deployment_config, tag=tag, color=color)

        self.wait_for_resource_running(
            'Deployment',
            'replicas',
            lambda deployment: deployment.replicas if deployment.ready and self.pod_find(tag, color) else None,
            self.deployment_timeout,
            *applied_objects
        )

        return color

    @log('Waiting for {resource_type} {prop} to become available')
    def wait_for_resource_running(self, resource_type, prop, prop_extractor, timeout_seconds, *object_names):
        def try_extract_value(resource_name):
            try:
                result = self.kube_api.resource_by_name(resource_type, resource_name)
                return prop_extractor(result or {})
            except (IndexError, KeyError, AttributeError):
                return None

        def extract_value_with_timeout(resource_name):
            value = None

            if timeout_seconds:
                attempts = 0

                while not value and attempts < timeout_seconds:
                    sleep(1)
                    attempts += 1
                    output('.', '', flush=True)
                    value = try_extract_value(resource_name)
            else:
                value = try_extract_value(resource_name)

            if value:
                output('\n{} {} {} is: {}'.format(resource_type, resource_name, prop, value))
            elif timeout_seconds:
                raise ActionFailedError(
                    '\nFailed after {} seconds elapsed. For more info try running: $ kubectl describe {} {}'.format(
                        timeout_seconds, resource_type, resource_name))

            return value

        values = [extract_value_with_timeout(name) for name in object_names]
        return values

    @log('Running smoke tests on {color} deployment...')
    def smoke_test(self, color):
        if self.smoke_service_config:
            def service_host_extractor(service):
                if self.is_minikube:
                    service_address = self.runner.start(cmd.MINIKUBE_SERVICE_URL.format(service.name), capture=True)
                else:
                    service_address = get_loadbalancer_address(service)

                return service_address if is_host_up(service_address) else None

            applied_objects = self.apply('smoke service', self.smoke_service_config, color=color)
            smoke_service_address = ','.join(self.wait_for_resource_running(
                'Service',
                'host',
                service_host_extractor,
                self.smoke_service_timeout,
                *applied_objects
            ))

            return_code = self.runner.start(self.smoke_tests_command, TEST_HOST=smoke_service_address, silent=True)
            return return_code == 0

        return True

    @log('Promoting {color} deployment...')
    def swap(self, color):
        self.apply('public service', self.service_config, color=color)
        self.wait_for_resource_running(
            'Service',
            'status',
            lambda service: 'ready' if service.exists(ensure=True) else None,
            self.service_timeout,
            self.service_name
        )

    @log('Publishing...')
    def publish(self):
        next_tag = self.build()
        self.push(next_tag)

        next_color = self.deploy(next_tag)
        is_initial, db_migrations_previous_state = self.migrate(next_tag, next_color)
        health_ok = self.smoke_test(next_color)

        if health_ok:
            self.swap(next_color)
        else:
            if not is_initial:
                self.migrate_rollback(next_tag, next_color, db_migrations_previous_state)

            raise ActionFailedError('Cannot promote {} deployment because smoke tests failed'.format(next_color))

        output('Done.')

    @log('Rolling back to previous deployment...')
    def rollback(self):
        color = self.other_env()

        if color:
            self.swap(color)
        else:
            raise ActionFailedError('Cannot rollback to a previous environment because one does not exist.')

        output('Done.')



from argparse import ArgumentParser
from traceback import format_exc

from bgkube.bg import BgKube
from bgkube.errors import BlueGreenError
from bgkube.utils import dot_env_dict, error


def get_parser():
    parser = ArgumentParser()
    parser.add_argument('command', help='', default='publish', choices=['publish', 'rollback', 'build', 'push'])
    parser.add_argument('command_args', nargs='*')
    parser.add_argument('-e', '--env-file', help='.env file for the options below and application vars in the configs')
    parser.add_argument('-c', '--cluster-name', help='unique name of the cluster')
    parser.add_argument('-z', '--cluster-zone', help='zone name of the cluster location')
    parser.add_argument('-m', '--docker-machine-name', help='name of the docker machine if applicable')
    parser.add_argument('--docker-build-args', help='arguments supplied to docker build command separated with spaces')
    parser.add_argument('-i', '--image-name', help='name of the container image to build using docker')
    parser.add_argument('-s', '--service-name', help='name of the main service intended to serve clients')
    parser.add_argument('--service-config', help='config of the main service')
    parser.add_argument('--service-timeout', help='timeout secs to wait for healthy state or return an error')
    parser.add_argument('--deployment-config', help='config of the deployment containing the main service pods')
    parser.add_argument('--deployment-timeout', help='timeout secs to wait for healthy state or return an error')
    parser.add_argument('-x', '--context', help='docker context path used to build the container image')
    parser.add_argument('-d', '--dockerfile', help='Dockerfile path')
    parser.add_argument('--smoke-service-config', help='config of the smoke service lb exposed for health checks')
    parser.add_argument('--smoke-tests-command', help='shell command to run health checks against the smoke service')
    parser.add_argument('--db-migrations-job-config-seed', help='job config to populate the database with initial data')
    parser.add_argument('--db-migrations-job-timeout', help='timeout secs to wait for healthy state or return an error')
    parser.add_argument(
        '--db-migrations-status-command',
        help='shell command executed on any of the running deployment pods to return the current migrations status'
    )
    parser.add_argument(
        '--db-migrations-apply-command',
        help='shell command executed on any of the running deployment pods to apply the latest migrations generated in '
             'the current image'
    )
    parser.add_argument(
        '--db-migrations-rollback-command',
        help='shell command executed on any of the running deployment pods with the migrations status command stdout '
             'as argument - retrieved before applying migrations, to perform a rollback to that state'
    )
    parser.add_argument('--kops-state-store', help='aws cluster state storage bucket name')
    parser.add_argument('--container-registry', help='container registry alias or implementation class')

    return parser


def load_options_from_env(options):
    for k, v in dict(dot_env_dict(options.env_file)).items():
        attr_name = k.lower()

        if hasattr(options, attr_name) and getattr(options, attr_name) is None:
            setattr(options, attr_name, v)


def run():
    parser = get_parser()
    options = parser.parse_args()

    if options.env_file:
        load_options_from_env(options)

    try:
        getattr(BgKube(options), options.command)(*options.command_args)
    except BlueGreenError as e:
        error(e)
    except Exception:
        error(format_exc())


if __name__ == '__main__':
    run()



from __future__ import print_function

import os
import sys
import time
import requests
from inspect import isclass

from bgkube.errors import RequiredOptionError


def require(obj, attr):
    value = getattr(obj, attr, None)

    if value is None:
        raise RequiredOptionError(attr)

    return value


def dot_env_dict(filename):
    # TODO: Handle lines with inline comment
    with open(filename) as fp:
        for line in fp.readlines():
            line = line.strip()

            if not line or line.startswith('#') or '=' not in line:
                continue

            k, v = line.split('=', 1)
            k, v = k.strip(), v.strip()

            quotes = ['\'', '"']
            v_unquoted = v.strip(''.join(quotes))

            if v_unquoted.isdigit() and len(v_unquoted) == len(v):
                v = int(v_unquoted)
            else:
                v = v_unquoted

            yield k, v


def read_with_merge_vars(filename, merge_vars):
    with open(filename) as fs:
        result = fs.read()

        for k in sorted(merge_vars.keys(), reverse=True):
            result = result.replace('${}'.format(k), str(merge_vars[k]))

        return result


def timestamp():
    return int(time.time())


def error(msg, **output_kwargs):
    output(msg, **output_kwargs)
    exit(1)


def output(msg, end='\n', **kwargs):
    if not os.environ.get('PYTEST_CURRENT_TEST', None):
        flush = kwargs.pop('flush', False)
        print(msg, end=end, **kwargs)
        if flush:
            sys.stdout.flush()


def log(message, **defaults):
    def wrap(func):
        def wrapped(*args, **kwargs):
            params = {}

            import inspect
            spec = inspect.getargspec(func)

            for i, v in enumerate(spec.args or ()):
                params[v] = args[i] if i < len(args) else kwargs.get(v, '')

            if 'self' in params:
                attrs = {k: v for k, v in params.pop('self').__dict__.items() if k[0].isalpha() and not callable(v)}
                params = dict(attrs, **params)

            for k, v in defaults.items():
                if not params.get(k, ''):
                    params[k] = v

            if message.strip().startswith('$'):
                output(message.format(**{k: '{}'.format(v) for k, v in params.items()}))
            else:
                output('=> ' + message.format(**{k: '\'{}\''.format(v) for k, v in params.items()}))
            result = func(*args, **kwargs)

            return result

        return wrapped

    return wrap


def module_type_subclasses(mod, t):
    def is_valid(obj):
        return isclass(obj) and issubclass(obj, t) and obj != t

    for name, cls in __import__(mod, fromlist=[mod]).__dict__.items():
        if is_valid(cls):
            yield name, cls


def get_loadbalancer_address(service):
    # TODO: Should be improved to either returning all possible addresses or get smarter at detecting the right one
    ports = service.obj['spec']['ports'][0]
    scheme = ports.get('name', 'http')
    port = ports['port']

    ingress = service.obj['status']['loadBalancer']['ingress'][0]
    host = ingress.get('ip', None) or ingress.get('hostname', None)

    return '{}://{}:{}'.format(scheme, host, port)


def is_host_up(address, status_code=200):
    try:
        response = requests.get(address)
        return response.status_code == status_code
    except requests.exceptions.ConnectionError:
        return False



from os import environ
from subprocess import Popen, PIPE

from bgkube.errors import ActionFailedError
from bgkube.utils import log


class Runner:
    REDIRECTION = '&> /dev/null'

    def __init__(self, *init_commands, **kwargs):
        self.init_commands = list(init_commands)

        runner = kwargs.get('runner', None)
        if runner:
            self.init_commands = runner.init_commands + self.init_commands

    def user_env(self, env):
        result = environ.copy()
        result.update(**env)

        return result

    def get_init_commands_silenced(self):
        def silence(command):
            return '{} {}'.format(command, self.REDIRECTION) if self.REDIRECTION not in command else command

        for i, _ in enumerate(self.init_commands):
            for separator in [' && ', ' || ']:
                self.init_commands[i] = separator.join(silence(p) for p in self.init_commands[i].split(separator))

        return '; '.join(silence(c) for c in self.init_commands)

    @log('$ {command}')
    def start(self, command, silent=False, capture=False, **env):
        def read(st_result):
            if st_result and isinstance(st_result, bytes):
                st_result = st_result.decode('utf-8').strip()
            return st_result

        target_command = '{}; {}'.format(self.get_init_commands_silenced(), command)
        kwargs = dict() if not capture else dict(stdin=PIPE, stdout=PIPE, stderr=PIPE)

        process = Popen(target_command, env=self.user_env(env), shell=True, **kwargs)
        output, error = process.communicate()
        return_code = process.returncode

        if return_code == 0 or silent:
            return return_code if not capture else read(output)
        else:
            raise ActionFailedError('Command invocation failed with code {}: {}'.format(return_code, read(error)))



# docker
DOCKER_BUILD = 'docker build {context} -f {dockerfile} -t {image}:{tag}'

# docker-machine
DOCKERMACHINE_ENV = 'which docker-machine && eval $(docker-machine env {})'

# gcloud
GCLOUD_CONTAINER_CLUSTER_GET_CREDENTIALS = 'gcloud container clusters get-credentials {cluster} --zone {zone}'
GCLOUD_DOCKER_PUSH = 'gcloud docker -- push {}'

# aws
AWS_ECR_GET_CREDENTIALS = '$(aws ecr get-login --no-include-email --region {region})'
AWS_KOPS_EXPORT_CONFIG = 'KOPS_STATE_STORE={store} kops export kubecfg --name {cluster}'
AWS_DOCKER_PUSH = 'docker push {}'

# kubectl
KUBECTL_EXEC = 'kubectl exec {pod} {command} {args}'
KUBECTL_CONFIG_MINIKUBE = 'kubectl config use-context minikube'

# minikube
MINIKUBE_SERVICE_URL = 'minikube service {} --url'
MINIKUBE_DOCKER_ENV = 'eval $(minikube docker-env)'



from bgkube import cmd
from bgkube.run import Runner
from bgkube.utils import module_type_subclasses


class ContainerRegistry:
    name = None

    def push(self, image):
        raise NotImplementedError()

    def __repr__(self):
        return self.name


class GoogleContainerRegistry(ContainerRegistry):
    alias = 'gcr'
    name = 'Google container registry'

    def __init__(self, runner, cluster_name, cluster_zone, **_):
        self.runner = Runner(
            cmd.GCLOUD_CONTAINER_CLUSTER_GET_CREDENTIALS.format(cluster=cluster_name, zone=cluster_zone),
            runner=runner
        )

    def push(self, image):
        self.runner.start(cmd.GCLOUD_DOCKER_PUSH.format(image))


class AwsContainerRegistry(ContainerRegistry):
    alias = 'ecr'
    name = 'Amazon EC2 Container Registry'

    def __init__(self, runner, cluster_name, cluster_zone, kops_state_store, **_):
        self.runner = Runner(
            cmd.AWS_ECR_GET_CREDENTIALS.format(region=cluster_zone),
            cmd.AWS_KOPS_EXPORT_CONFIG.format(cluster=cluster_name, store=kops_state_store),
            runner=runner
        )

    def push(self, image):
        self.runner.start(cmd.AWS_DOCKER_PUSH.format(image))


class LocalContainerRegistry(ContainerRegistry):
    alias = 'local'
    name = 'Local container registry'

    def __init__(self, runner, **_):
        self.runner = Runner(
            cmd.KUBECTL_CONFIG_MINIKUBE,
            runner=runner
        )

    def push(self, image):
        pass


DEFAULT = 'gcr'
MODULE = __name__
BUILTIN = {cls.alias: '{}.{}'.format(MODULE, name) for name, cls in module_type_subclasses(MODULE, ContainerRegistry)}


def load(runner, options):
    kwargs = options if isinstance(options, dict) else vars(options)
    container_registry = kwargs.get('container_registry', None)

    if container_registry in BUILTIN or not container_registry:
        selected_registry = BUILTIN[container_registry or DEFAULT]
    else:
        selected_registry = container_registry

    module_name, class_name = '.'.join(selected_registry.split('.')[:-1]), selected_registry.split('.')[-1]
    reg = [cls for name, cls in module_type_subclasses(module_name, ContainerRegistry) if name == class_name][0]

    return reg(runner, **kwargs)



from distutils.core import setup
from pip.req import parse_requirements

install_req = parse_requirements('requirements/core.txt', session='skip')
req = [str(ir.req) for ir in install_req]


def read(filename):
    with open(filename) as file:
        return file.read()


setup(
    name='bg-kube',
    packages=['bgkube'],
    version='0.0.13',
    description='An interface for automating blue-green deployments using Kubernetes',
    long_description=read('README.rst'),
    author='Phivos Stylianides',
    author_email='stphivos@gmail.com',
    url='https://github.com/stphivos/bg-kube',
    keywords='Kubernetes microservices blue-green GKE AWS Kops CI/CD Continuous integration delivery deployment',
    classifiers=[],
    install_requires=req,
    entry_points={'console_scripts': ['bg-kube = bgkube.__main__:run', ], },
)



class BlueGreenError(Exception):
    pass


class RequiredOptionError(BlueGreenError):
    def __init__(self, *args):
        super(RequiredOptionError, self).__init__('Option \'{}\' is required and was not supplied'.format(*args))


class InvalidOptionValueError(BlueGreenError):
    def __init__(self, *args):
        super(InvalidOptionValueError, self).__init__('Invalid value \'{}\' supplied for option \'{}\''.format(*args))


class ActionFailedError(BlueGreenError):
    pass
